/*
 * Copyright 2019 Blender Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __JITTER_H__
#define __JITTER_H__

#include "util/types.h"

CCL_NAMESPACE_BEGIN

void progressive_multi_jitter_generate_2D(float2 points[], int size, int rng_seed);
void progressive_multi_jitter_02_generate_2D(float2 points[], int size, int rng_seed);

CCL_NAMESPACE_END

#endif /* __JITTER_H__ */
